indexing
	description: "A simple cache where items can be recalled %
                %from storage on demand"
	license: "Eiffel Forumn License v2 - See license.txt"
	source: "{
		Amalasoft Corporation
		273 Harwood Avenue
		Littleton, MA 01460 USA
		}"
	status: "Copyright 1995-2008, Amalasoft";
	date: "$Date: 2008/12/03 $";
	revision: "$Revision: 001$";
	howto: "{
    To use this class, create a new instance with a given number of items
}"

class AEL_DS_CACHE [G -> AEL_DS_CACHEABLE create make_with_cache_key end]

create
	make

--|========================================================================
feature {NONE} -- Initialization
--|========================================================================

	make (v: INTEGER) is
			-- Create Current with a capacity of 'v'
		do
			create queue.make
			queue.start
			capacity := v
		end

--|========================================================================
feature -- Element change
--|========================================================================

	retrieve_item_by_key (k: HASHABLE) is
			-- Retrieve the item identified by the key 'k'
			-- Fetch from cache if available; retrieve from storage if not
			--
			-- Move to back of LRU queue (making it most recently used)
			--
			-- Leave retrieved item in 'retrieved_item'
		require
			exists: k /= Void
		do
			retrieved_item := cached_item_by_key (k)
			if retrieved_item = Void then
				create retrieved_item.make_with_cache_key (k)
			else
				remove_item_from_queue (retrieved_item)
			end
			queue.extend (retrieved_item)
		ensure
			retrieved_item /= Void implies retrieved_item.cache_key.is_equal (k)
		end

	--|--------------------------------------------------------------

	force_retrieve_item_by_key (k: HASHABLE) is
			-- Remove from the LRU queue the item identified by the key 'k'
			-- Then retrieve the item as usual
		require
			exists: k /= Void
		do
			if has_item_by_key (k) then
				remove_item_from_queue_by_key (k)
			end
			retrieve_item_by_key (k)
		ensure
			retrieved_item /= Void implies retrieved_item.cache_key.is_equal (k)
		end

--|========================================================================
feature -- Access
--|========================================================================

	retrieved_item: G
			-- Item most recently retrieved

	--|--------------------------------------------------------------

	cached_item_by_key (k: HASHABLE): like retrieved_item is
			-- Item by key, if in cache (not retrieved automatically)
		require
			exists: k /= Void
		do
			from queue.start
			until queue.after or Result /= Void
			loop
				if queue.item.cache_key.is_equal (k) then
					Result := queue.item
				else
					queue.forth
				end
			end
			queue.start
		ensure
			exists: has_item_by_key (k) implies Result /= Void
		end

--|========================================================================
feature -- Status
--|========================================================================

	capacity: INTEGER
			-- Maximum number of items in Current at any given time

	count: INTEGER is
			-- Number of items in Current presently
		do
			Result := queue.count
		end

 --|------------------------------------------------------------------------

	has_item_by_key (k: HASHABLE): BOOLEAN is
			-- Does Current contain, in cache, the given item
		require
			exists: k /= Void
		do
			from queue.start
			until queue.after or Result
			loop
				if queue.item.cache_key.is_equal (k) then
					Result := True
				else
					queue.forth
				end
			end
			queue.start
		end

--|========================================================================
feature {NONE} -- Implementation
--|========================================================================

	queue: TWO_WAY_LIST [G]
			-- Queue of items for LRU logic
			-- Front of queue is least recently accessed item

	--|--------------------------------------------------------------

	remove_item_from_queue (v: like retrieved_item) is
			-- Remove the given item from the LRU queue
		require
			exists: v /= Void
		do
			queue.start
			queue.search (v)
			if not queue.exhausted then
				queue.remove
			end
			queue.start
		end

	--|--------------------------------------------------------------

	remove_item_from_queue_by_key (k: HASHABLE) is
			-- Remove from the LRU queue, the item identified by 'k'
		require
			exists: k /= Void
		local
			found: BOOLEAN
		do
			from queue.start
			until queue.exhausted or found
			loop
				if queue.item.cache_key.is_equal (k) then
					queue.remove
					found := True
				else
					queue.forth
				end
			end
			queue.start
		end

 --|------------------------------------------------------------------------
invariant
	queue_exists: queue /= Void
	valid_queue_state:  queue.is_empty implies queue.after
	bounded: count <= capacity

end -- class AEL_DS_CACHE
